"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = __importStar(require("path"));
const restrict_imports_loader_1 = require("restrict-imports-loader");
const terser_webpack_plugin_1 = __importDefault(require("terser-webpack-plugin"));
const tsconfig_paths_webpack_plugin_1 = __importDefault(require("tsconfig-paths-webpack-plugin"));
const Metadata = __importStar(require("userscript-metadata"));
const webpack_userscript_1 = __importDefault(require("webpack-userscript"));
const configuration_1 = require("./configuration");
const mode_1 = require("./mode");
const sass_1 = require("./sass");
const utilities_1 = require("./utilities");
const webpack_plugin_1 = require("./webpack-plugin");
const USERSCRIPTER_BUILD = "userscripter/build";
const EXTENSIONS = {
    TS: ["ts", "tsx"],
    JS: ["mjs", "js", "jsx"],
    SASS: ["scss"],
    SVG: ["svg"],
};
const WEBPACK_STATS_TO_STRING_OPTIONS = {
    depth: false,
    hash: false,
    modules: false,
    entrypoints: false,
    colors: true,
};
exports.DEFAULT_BUILD_CONFIG = x => ({
    allowJs: false,
    id: x.id,
    hostedAt: null,
    mainFile: "main.ts",
    mode: mode_1.Mode.development,
    nightly: false,
    now: x.now,
    outDir: "dist",
    rootDir: x.rootDir,
    sassVariableGetter: "getGlobal",
    sassVariables: {},
    sourceDir: "src",
    verbose: false,
});
exports.DEFAULT_METADATA_SCHEMA = {
    items: {
        ...Metadata.DEFAULT_ITEMS,
        version: Metadata.DEFAULT_ITEMS.version.butRequired(),
        run_at: Metadata.DEFAULT_ITEMS.run_at.butRequired(),
    },
    warnings: Metadata.DEFAULT_WARNINGS,
    underscoresAsHyphens: true,
};
function createWebpackConfig(x) {
    const buildConfig = (configuration_1.overrideBuildConfig(x.buildConfig, x.env).buildConfig);
    const { allowJs, id, mainFile, mode, nightly, now, outDir, rootDir, sassVariableGetter, sassVariables, sourceDir, verbose, } = buildConfig;
    if (verbose) {
        const envVarLines = configuration_1.envVars(x.env).map(e => "  " + e[0] + ": " + (e[1] === undefined ? "(not specified)" : e[1]));
        console.log("Environment variables:");
        console.log(envVarLines.join("\n"));
        console.log();
        console.log("Effective build config (after considering environment variables):");
        console.log(buildConfig);
        console.log();
    }
    const metadata = x.metadata(buildConfig);
    return {
        mode: mode,
        entry: {
            userscript: resolveIn(sourceDir)(mainFile),
        },
        output: {
            path: resolveIn(rootDir)(outDir),
            filename: configuration_1.distFileName(id, "user"),
        },
        stats: WEBPACK_STATS_TO_STRING_OPTIONS,
        module: {
            rules: [
                {
                    test: filenameExtensionRegex(EXTENSIONS.SVG),
                    loaders: [
                        {
                            loader: require.resolve("raw-loader"),
                        },
                    ],
                },
                {
                    test: filenameExtensionRegex(EXTENSIONS.SASS),
                    loaders: [
                        {
                            loader: require.resolve("to-string-loader"),
                        },
                        {
                            loader: require.resolve("css-loader"),
                            options: {
                                sourceMap: false,
                                modules: {
                                    localIdentName: "[local]",
                                },
                            },
                        },
                        {
                            loader: require.resolve("sass-loader"),
                            options: {
                                sassOptions: {
                                    functions: { [sassVariableGetter]: sass_1.getGlobalFrom(sassVariables) },
                                },
                            },
                        },
                    ],
                },
                {
                    test: filenameExtensionRegex(EXTENSIONS.TS),
                    include: resolveIn(rootDir)(sourceDir),
                    loaders: [
                        {
                            loader: require.resolve("awesome-typescript-loader"),
                            options: { allowJs, useCache: true },
                        },
                        {
                            loader: require.resolve("restrict-imports-loader"),
                            options: {
                                severity: "error",
                                detailedErrorMessages: true,
                                rules: [
                                    {
                                        restricted: restrict_imports_loader_1.everythingInPackage(USERSCRIPTER_BUILD),
                                        severity: "fatal",
                                        info: `"${USERSCRIPTER_BUILD}" and its submodules cannot be imported in the source directory ('${sourceDir}'). Please remove these imports:`,
                                    },
                                ],
                            },
                        },
                    ],
                },
            ],
        },
        resolve: {
            plugins: [
                new tsconfig_paths_webpack_plugin_1.default(),
            ],
            extensions: utilities_1.concat(Object.values(EXTENSIONS)).map(e => "." + e),
        },
        plugins: [
            new webpack_plugin_1.UserscripterWebpackPlugin(x),
            new webpack_userscript_1.default({
                metajs: true,
                headers: {
                    ...metadata,
                    name: metadata.name + (nightly ? " Nightly" : ""),
                    version: (nightly || mode === mode_1.Mode.development
                        ? metadata.version + "." + dateAsSemver(now)
                        : metadata.version),
                },
            }),
        ],
        optimization: {
            minimize: mode === mode_1.Mode.production,
            minimizer: [
                new terser_webpack_plugin_1.default({
                    parallel: true,
                }),
            ],
        },
    };
}
exports.createWebpackConfig = createWebpackConfig;
const resolveIn = (root) => (subdir) => path.resolve(root, subdir);
function filenameExtensionRegex(extensions) {
    return new RegExp("\\.(" + extensions.join("|") + ")$");
}
function dateAsSemver(d) {
    return [
        d.getFullYear(),
        d.getMonth() + 1,
        d.getDate(),
        d.getHours(),
        d.getMinutes(),
    ].join(".");
}
