import * as path from "path";
import { everythingInPackage } from "restrict-imports-loader";
import TerserPlugin from "terser-webpack-plugin";
import TsconfigPathsPlugin from "tsconfig-paths-webpack-plugin";
import * as Metadata from "userscript-metadata";
import WebpackUserscriptPlugin from "webpack-userscript";
import { distFileName, envVars, overrideBuildConfig, } from "./configuration";
import { Mode } from "./mode";
import { getGlobalFrom } from "./sass";
import { concat } from "./utilities";
import { UserscripterWebpackPlugin } from "./webpack-plugin";
const USERSCRIPTER_BUILD = "userscripter/build";
const EXTENSIONS = {
    TS: ["ts", "tsx"],
    JS: ["mjs", "js", "jsx"],
    SASS: ["scss"],
    SVG: ["svg"],
};
const WEBPACK_STATS_TO_STRING_OPTIONS = {
    depth: false,
    hash: false,
    modules: false,
    entrypoints: false,
    colors: true,
};
export const DEFAULT_BUILD_CONFIG = x => ({
    allowJs: false,
    id: x.id,
    hostedAt: null,
    mainFile: "main.ts",
    mode: Mode.development,
    nightly: false,
    now: x.now,
    outDir: "dist",
    rootDir: x.rootDir,
    sassVariableGetter: "getGlobal",
    sassVariables: {},
    sourceDir: "src",
    verbose: false,
});
export const DEFAULT_METADATA_SCHEMA = {
    items: {
        ...Metadata.DEFAULT_ITEMS,
        version: Metadata.DEFAULT_ITEMS.version.butRequired(),
        run_at: Metadata.DEFAULT_ITEMS.run_at.butRequired(),
    },
    warnings: Metadata.DEFAULT_WARNINGS,
    underscoresAsHyphens: true,
};
export function createWebpackConfig(x) {
    const buildConfig = (overrideBuildConfig(x.buildConfig, x.env).buildConfig);
    const { allowJs, id, mainFile, mode, nightly, now, outDir, rootDir, sassVariableGetter, sassVariables, sourceDir, verbose, } = buildConfig;
    if (verbose) {
        const envVarLines = envVars(x.env).map(e => "  " + e[0] + ": " + (e[1] === undefined ? "(not specified)" : e[1]));
        console.log("Environment variables:");
        console.log(envVarLines.join("\n"));
        console.log();
        console.log("Effective build config (after considering environment variables):");
        console.log(buildConfig);
        console.log();
    }
    const metadata = x.metadata(buildConfig);
    return {
        mode: mode,
        entry: {
            userscript: resolveIn(sourceDir)(mainFile),
        },
        output: {
            path: resolveIn(rootDir)(outDir),
            filename: distFileName(id, "user"),
        },
        stats: WEBPACK_STATS_TO_STRING_OPTIONS,
        module: {
            rules: [
                {
                    test: filenameExtensionRegex(EXTENSIONS.SVG),
                    loaders: [
                        {
                            loader: require.resolve("raw-loader"),
                        },
                    ],
                },
                {
                    test: filenameExtensionRegex(EXTENSIONS.SASS),
                    loaders: [
                        {
                            loader: require.resolve("to-string-loader"),
                        },
                        {
                            loader: require.resolve("css-loader"),
                            options: {
                                sourceMap: false,
                                modules: {
                                    localIdentName: "[local]",
                                },
                            },
                        },
                        {
                            loader: require.resolve("sass-loader"),
                            options: {
                                sassOptions: {
                                    functions: { [sassVariableGetter]: getGlobalFrom(sassVariables) },
                                },
                            },
                        },
                    ],
                },
                {
                    test: filenameExtensionRegex(EXTENSIONS.TS),
                    include: resolveIn(rootDir)(sourceDir),
                    loaders: [
                        {
                            loader: require.resolve("awesome-typescript-loader"),
                            options: { allowJs, useCache: true },
                        },
                        {
                            loader: require.resolve("restrict-imports-loader"),
                            options: {
                                severity: "error",
                                detailedErrorMessages: true,
                                rules: [
                                    {
                                        restricted: everythingInPackage(USERSCRIPTER_BUILD),
                                        severity: "fatal",
                                        info: `"${USERSCRIPTER_BUILD}" and its submodules cannot be imported in the source directory ('${sourceDir}'). Please remove these imports:`,
                                    },
                                ],
                            },
                        },
                    ],
                },
            ],
        },
        resolve: {
            plugins: [
                new TsconfigPathsPlugin(),
            ],
            extensions: concat(Object.values(EXTENSIONS)).map(e => "." + e),
        },
        plugins: [
            new UserscripterWebpackPlugin(x),
            new WebpackUserscriptPlugin({
                metajs: true,
                headers: {
                    ...metadata,
                    name: metadata.name + (nightly ? " Nightly" : ""),
                    version: (nightly || mode === Mode.development
                        ? metadata.version + "." + dateAsSemver(now)
                        : metadata.version),
                },
            }),
        ],
        optimization: {
            minimize: mode === Mode.production,
            minimizer: [
                new TerserPlugin({
                    parallel: true,
                }),
            ],
        },
    };
}
const resolveIn = (root) => (subdir) => path.resolve(root, subdir);
function filenameExtensionRegex(extensions) {
    return new RegExp("\\.(" + extensions.join("|") + ")$");
}
function dateAsSemver(d) {
    return [
        d.getFullYear(),
        d.getMonth() + 1,
        d.getDate(),
        d.getHours(),
        d.getMinutes(),
    ].join(".");
}
